import logging
import asyncio
import functools
from contextlib import contextmanager


logger = logging.getLogger('cterasdk.core')


def update_current_tenant(execute_request):
    if asyncio.iscoroutinefunction(execute_request):
        @functools.wraps(execute_request)
        async def execute_and_update(self, tenant):
            with update_current_session(self, tenant):
                response = await execute_request(self, tenant)
            return response
    else:
        @functools.wraps(execute_request)
        def execute_and_update(self, tenant):
            with update_current_session(self, tenant):
                response = execute_request(self, tenant)
            return response
    return execute_and_update


@contextmanager
def update_current_session(core, tenant):
    yield
    tenant = tenant if tenant else 'Administration'
    logger.debug('Updating current session. Tenant: %s', tenant)
    core.session().update_current_tenant(tenant)
    logger.debug('Updated current session. Tenant: %s', tenant)



import logging
from ..exceptions.direct import BlockInfo


logger = logging.getLogger('cterasdk.direct')


def blocks(file, array):
    """
    Filter Blocks by Block Number.

    :param list[cterasdk.direct.types.File] file: File Object.
    :param list[cterasdk.direct.types.BlockInfo] array: List of BlockInfo objects,
     or list of integers identifying the block position.
    :returns: List of Chunks.
    :rtype: list[cterasdk.direct.types.Chunk]
    """
    if array is not None:
        numbers = [block.number if isinstance(block, BlockInfo) else block for block in array]
        return [file.chunks[number - 1] for number in numbers]
    return file.chunks


def span(file, byte_range):
    """
    Filter Blocks by Byte Range.

    :param list[cterasdk.direct.types.File] file: File Object.
    :param list[cterasdk.direct.types.ByteRange] byte_range: Byte Range.
    :returns: List of Chunks.
    :rtype: list[cterasdk.direct.types.Chunk]
    """
    if validate_byte_range(file, byte_range):
        start, end = None, None
        for index, chunk in enumerate(file.chunks):
            if byte_range.start < chunk.offset + chunk.length:
                start = index
                break
        if not byte_range.eof:
            for index, chunk in enumerate(file.chunks):
                if byte_range.end < chunk.offset + chunk.length:
                    end = index + 1
                    break
        return file.chunks[start:end]
    return file.chunks


def validate_byte_range(file, byte_range):
    size = file.size
    if byte_range.start > size:
        logger.error('Byte range start %s is greater than the file size %s.', byte_range.start, size)
        return False
    if not byte_range.eof and byte_range.end >= size:
        logger.debug('Byte range end is greater than file size.')
        byte_range.eof = True
    return True



import logging

from .base_command import BaseCommand
from .types import UserAccount
from ..exceptions import CTERAException, ObjectNotFoundException
from ..exceptions.session import ContextError
from ..common import Object, DateTimeUtils
from ..common import union
from . import query


logger = logging.getLogger('cterasdk.core')


class Users(BaseCommand):
    """
    Portal User Management APIs

    :ivar cterasdk.core.credentials.Credentials credentials: Object holding Portal User Credential Management APIs.
    """

    default = ['name']

    def __init__(self, portal):
        super().__init__(portal)
        self.credentials = Credentials(self._core)

    def _get_entire_object(self, user_account):
        ref = f'/users/{user_account.name}' if user_account.is_local \
            else f'/domains/{user_account.directory}/adUsers/{user_account.name}'
        try:
            return self._core.api.get(ref)
        except CTERAException as error:
            raise CTERAException(f'Could not retrieve user: {ref}') from error

    def get(self, user_account, include=None):
        """
        Get a user account

        :param cterasdk.core.types.UserAccount user_account: User account, including the user directory and user name
        :param list[str] include: List of fields to retrieve, defaults to ['name']
        :return: The user account, including the requested fields
        """
        baseurl = f'/users/{user_account.name}' if user_account.is_local \
            else f'/domains/{user_account.directory}/adUsers/{user_account.name}'
        include = union(include or [], Users.default)
        include = ['/' + attr for attr in include]
        user_object = self._core.api.get_multi(baseurl, include)
        if user_object.name is None:
            raise ObjectNotFoundException(baseurl)
        return user_object

    def list_local_users(self, include=None):
        """
        List all local users

        :param list[str] include: List of fields to retrieve, defaults to ['name']
        :return: Iterator for all local users
        :rtype: cterasdk.lib.iterator.Iterator
        """
        include = union(include or [], Users.default)
        param = query.QueryParamBuilder().include(include).build()
        return query.iterator(self._core, '/users', param)

    def list_domain_users(self, domain, include=None):
        """
        List all the users in the domain

        :param str domain: Domain name
        :param list[str] include: List of fields to retrieve, defaults to ['name']
        :return: Iterator for all the domain users
        :rtype: cterasdk.lib.iterator.Iterator
        """
        include = union(include or [], Users.default)
        param = query.QueryParamBuilder().include(include).build()
        return query.iterator(self._core, f'/domains/{domain}/adUsers', param)

    def add(self, name, email, first_name, last_name, password, role, company=None, comment=None, password_change=False):
        """
        Create a local user account

        :param str name: User name for the new user
        :param str email: E-mail address of the new user
        :param str first_name: The first name of the new user
        :param str last_name: The last name of the new user
        :param str password: Password for the new user
        :param cterasdk.core.enum.Role role: User role of the new user
        :param str,optional company: The name of the company of the new user, defaults to None
        :param str,optional comment: Additional comment for the new user, defaults to None
        :param variable,optional password_change:
            Require the user to change the password on the first login.
            Pass datetime.date for a specific date, integer for days from creation, or True for immediate , defaults to False
        """
        param = Object()
        param._classname = "PortalUser"  # pylint: disable=protected-access
        param.name = name
        param.email = email
        param.firstName = first_name
        param.lastName = last_name
        param.password = password
        param.role = role
        param.company = company
        param.comment = comment
        if password_change:
            param.requirePasswordChangeOn = DateTimeUtils.get_expiration_date(password_change).strftime('%Y-%m-%d')

        logger.info('Creating user. %s', {'user': name})
        response = self._core.api.add('/users', param)
        logger.info('User created. %s', {'user': name, 'email': email, 'role': role})

        return response

    def modify(self, current_username, new_username=None, email=None, first_name=None,
               last_name=None, password=None, role=None, company=None, comment=None):
        """
        Modify a local user account

        :param str current_username: The current user name
        :param str,optional new_username: New name
        :param str,optional email: E-mail address
        :param str,optional first_name: First name
        :param str,optional last_name: Last name
        :param str,optional password: Password
        :param cterasdk.core.enum.Role,optional role: User role
        :param str,optional company: Company name
        :param str,optional comment: Comment
        """
        user_account = UserAccount(current_username)
        user = self._get_entire_object(user_account)
        if new_username:
            user.name = new_username
        if email:
            user.email = email
        if first_name:
            user.firstName = first_name
        if last_name:
            user.lastName = last_name
        if password:
            user.password = password
        if role:
            user.role = role
        if company is not None:
            user.company = company
        if comment is not None:
            user.comment = comment

        ref = f'/users/{current_username}'
        try:
            response = self._core.api.put(ref, user)
            logger.info("User modified. %s", {'username': user.name})
            return response
        except CTERAException as error:
            logger.error('User modification failed: %s', ref)
            raise CTERAException(f'User modification failed: {ref}') from error

    def apply_changes(self, wait=False):
        """
        Apply provisioning changes.\n

        :param bool,optional wait: Wait for all changes to apply
        """
        param = Object()
        param.objectId = None
        param.type = 'users'
        logger.info('Applying provisioning changes.')
        task = self._core.api.execute('', 'updateAccounts', param)
        if wait:
            task = self._core.tasks.wait(task)
        return task

    def delete(self, user):
        """
        Delete a user

        :param cterasdk.core.types.UserAccount user: the user account
        """
        logger.info('Deleting user. %s', {'user': str(user)})
        baseurl = f'/users/{user.name}' if user.is_local else f'/domains/{user.directory}/adUsers/{user.name}'
        response = self._core.api.execute(baseurl, 'delete', True)
        logger.info('User deleted. %s', {'user': str(user)})

        return response

    def generate_ticket(self, username, tenant):
        """
        Create an SSO ticket for a user.

        :param str username: Username
        :param str portal: Tenant
        """
        if self.session().in_tenant_context():
            raise ContextError('Browse the Global Administration to invoke this API')
        param = Object()
        param.username = username
        param.portal = tenant
        return self._core.api.execute('', 'getSessionToken', param)


class Credentials(BaseCommand):
    """
    Portal Credential Management APIs

    :ivar cterasdk.core.credentials.S3 s3: Object holding the Portal User S3 Credential Management APIs.
    """

    def __init__(self, portal):
        super().__init__(portal)
        self.s3 = S3(self._core)


class S3(BaseCommand):
    """
    S3 Credential Management APIs
    """

    def all(self, user_account):
        """
        List Credentials

        :param cterasdk.core.types.UserAccount user_account: User account
        """
        return self._core.credentials.s3.all(user_account)

    def create(self, user_account):
        """
        Create Credential

        :param cterasdk.core.types.UserAccount user_account: User account
        """
        return self._core.credentials.s3.create(user_account)

    def delete(self, access_key_id, user_account):
        """
        Delete Credential

        :param str access_key_id: Access Key ID
        :param cterasdk.core.types.UserAccount user_account: User account
        """
        return self._core.credentials.s3.delete(access_key_id, user_account)



import re
import logging
from contextlib import contextmanager
from ..objects.uri import quote, unquote
from ..common import Object, DateTimeUtils
from ..core.enum import ProtectionLevel, CollaboratorType, SearchType, PortalAccountType, FileAccessMode
from ..core.types import PortalAccount, UserAccount, GroupAccount
from ..exceptions.io import ResourceExistsError, PathValidationError, NameSyntaxError, ReservedNameError
from . import common


logger = logging.getLogger('cterasdk.core')


class CorePath(common.BasePath):
    """Path for CTERA Portal"""

    def __init__(self, scope, reference):
        """
        Initialize a CTERA Portal Path.

        :param str scope: Scope.
        :param str reference: Reference.
        """
        if isinstance(reference, Object):
            super().__init__(*CorePath._from_server_object(reference))
        elif isinstance(reference, str):
            super().__init__(scope, reference)
        elif reference is None:
            super().__init__(scope, '')
        else:
            message = 'Path validation failed: ensure the path exists and is correctly formatted.'
            logger.error(message)
            raise ValueError(message)

    @staticmethod
    def _from_server_object(reference):
        """
        Parse Path from Server Object.

        :param object reference: Path.
        :returns: Base Path and Relative Path
        :rtype: tuple(str)
        """
        classname = reference.__dict__.get('_classname', None)

        href = None
        if classname == 'ResourceInfo':
            href = reference.href
        elif classname == 'SnapshotResp':
            href = f'{reference.url}{reference.path}'
        else:
            raise ValueError('Could not determine server object: {classname}')

        href = unquote(href)
        match = re.search('^/?(ServicesPortal|admin)/webdav', href)
        start, end = match.span()
        return (href[start: end], href[end + 1:])

    @property
    def absolute(self):
        reference = self.reference.as_posix()
        previous_versions = 'PreviousVersions/'
        if previous_versions in reference:
            index = reference.index(previous_versions) + len(previous_versions)
            return f'{self.scope.as_posix()}/{quote(reference[:index]) + reference[index:]}'
        return super().absolute

    @staticmethod
    def instance(scope, entries):
        if isinstance(entries, list):
            return [CorePath(scope, e) for e in entries]
        return CorePath(scope, entries)


class SrcDstParam(Object):

    __instance = None

    @staticmethod
    def instance(src, dest=None):
        SrcDstParam(src, dest)
        return SrcDstParam.__instance

    def __init__(self, src, dest=None):
        super().__init__()
        self._classname = self.__class__.__name__
        self.src = src
        self.dest = dest
        SrcDstParam.__instance = self  # pylint: disable=unused-private-member


class ActionResourcesParam(Object):

    __instance = None

    @staticmethod
    def instance():
        ActionResourcesParam()
        return ActionResourcesParam.__instance

    def __init__(self):
        super().__init__()
        self._classname = self.__class__.__name__
        self.urls = []
        ActionResourcesParam.__instance = self  # pylint: disable=unused-private-member

    def add(self, param):
        self.urls.append(param)


class CreateShareParam(Object):

    __instance = None

    @staticmethod
    def instance(path, access, expire_on):
        CreateShareParam(path, access, expire_on)
        return CreateShareParam.__instance

    def __init__(self, path, access, expire_on):
        super().__init__()
        self._classname = self.__class__.__name__
        self.url = path
        self.share = Object()
        self.share._classname = 'ShareConfig'
        self.share.accessMode = access
        self.share.protectionLevel = 'publicLink'
        self.share.expiration = expire_on
        self.share.invitee = Object()
        self.share.invitee._classname = 'Collaborator'
        self.share.invitee.type = 'external'
        CreateShareParam.__instance = self  # pylint: disable=unused-private-member


class FetchResourcesParam(Object):

    def __init__(self):
        super().__init__()
        self._classname = 'FetchResourcesParam'
        self.start = 0
        self.limit = 100

    def increment(self):
        self.start = self.start + self.limit


class FetchResourcesParamBuilder:

    def __init__(self):
        self.param = FetchResourcesParam()

    def root(self, root):
        self.param.root = root  # pylint: disable=attribute-defined-outside-init
        return self

    def depth(self, depth):
        self.param.depth = depth  # pylint: disable=attribute-defined-outside-init
        return self

    def searchCriteria(self, criteria):
        self.param.searchCriteria = criteria  # pylint: disable=attribute-defined-outside-init
        return self

    def include_deleted(self):
        self.param.includeDeleted = True  # pylint: disable=attribute-defined-outside-init
        return self

    def limit(self, limit):
        self.param.limit = limit
        return self

    def build(self):
        return self.param


@contextmanager
def fetch_resources(path, depth, include_deleted, search_criteria, limit):
    """
    Create List Directory Parameter.
    """
    depth = depth if depth is not None else 1
    builder = FetchResourcesParamBuilder().root(path.absolute_encode).depth(depth)
    if include_deleted:
        builder.include_deleted()
    if search_criteria:
        builder.searchCriteria(search_criteria)
    if limit:
        builder.limit(limit)
    logger.info('Listing directory: %s', path.reference.as_posix())
    yield builder.build()


@contextmanager
def versions(path):
    logger.info('Listing versions: %s', path.reference.as_posix())
    yield


@contextmanager
def makedir(path):
    param = Object()
    param.name = path.name
    param.parentPath = path.parent.absolute_encode
    logger.info('Creating directory: %s', path.reference.as_posix())
    yield param
    logger.info('Directory created: %s', path.reference.as_posix())


@contextmanager
def rename(path, name):
    param = ActionResourcesParam.instance()
    param.add(SrcDstParam.instance(src=path.absolute, dest=path.parent.join(name).absolute))
    logger.info('Renaming item: %s to: %s', path.reference.as_posix(), name)
    yield param


def _delete_or_recover(paths, *, message=None):
    param = ActionResourcesParam.instance()
    paths = [paths] if not isinstance(paths, list) else paths
    for path in paths:
        param.add(SrcDstParam.instance(src=path.absolute))
        if message:
            logger.info('%s: %s', message, path.reference.as_posix())
    yield param


@contextmanager
def delete(*paths):
    return _delete_or_recover(list(paths), message='Deleting item')


@contextmanager
def recover(*paths):
    return _delete_or_recover(list(paths), message='Recovering item')


def _copy_or_move(paths, destination, *, message=None):
    param = ActionResourcesParam.instance()
    paths = [paths] if not isinstance(paths, list) else paths
    for path in paths:
        param.add(SrcDstParam.instance(src=path.absolute, dest=destination.join(path.name).absolute))
        if message:
            logger.info('%s from: %s to: %s', message, path.reference.as_posix(), destination.join(path.name).reference.as_posix())
    yield param


@contextmanager
def copy(*paths, destination=None):
    return _copy_or_move(list(paths), destination, message='Copying item')


@contextmanager
def move(*paths, destination=None):
    return _copy_or_move(list(paths), destination, message='Moving item')


@contextmanager
def public_link(path, access, expire_in):
    access = {'RO': 'ReadOnly', 'RW': 'ReadWrite', 'PO': 'PreviewOnly'}.get(access)
    expire_on = DateTimeUtils.get_expiration_date(expire_in).strftime('%Y-%m-%d')
    logger.info('Creating Public Link for: %s. Access: %s. Expires: %s', path.reference.as_posix(), access, expire_on)
    param = CreateShareParam.instance(path=path.absolute, access=access, expire_on=expire_on)
    yield param


@contextmanager
def handle(path):
    logger.info('Getting file handle: %s', path.reference)
    yield path.reference


@contextmanager
def upload(core, name, destination, size, fd):
    fd, size = common.encode_stream(fd, size)
    param = dict(
        name=name,
        Filename=name,
        fullpath=core.io.builder(CorePath(destination.reference, name).absolute_encode),
        fileSize=size,
        file=fd
    )
    logger.info('Uploading: %s to: %s', name, destination.reference)
    yield param


@contextmanager
def handle_many(directory, objects):
    param = Object()
    param.paths = ['/'.join([directory.absolute, filename]) for filename in objects]
    param.snapshot = None
    param.password = None
    param.portalName = None
    param.showDeleted = False
    logger.info('Getting directory handle: %s', directory.reference)
    yield param


@contextmanager
def share_info(path):
    logger.info('Getting Share info: %s', path.reference.as_posix())
    yield path.absolute_encode


def _is_sharing_on_user_behalf(path):
    return re.match(r'^Users/[^/]+(?=/)', path.reference.as_posix()) is not None


def _is_subfolder(path):
    depth = len(path.reference.parts)
    sharing_on_user_behalf = _is_sharing_on_user_behalf(path)
    return (sharing_on_user_behalf and depth > 3) or (not sharing_on_user_behalf and depth > 1)


@contextmanager
def share(path, as_project, allow_reshare, allow_sync, shares=None):
    if _is_subfolder(path):
        as_project = False  # Can't be a team project
        allow_sync = False  # Can't allow sync of a sub-folder

    param = Object()
    param._classname = 'ShareResourceParam'  # pylint: disable=protected-access
    param.url = path.absolute
    param.teamProject = as_project
    param.allowReshare = allow_reshare
    param.shouldSync = allow_sync
    param.shares = []
    if shares is not None:
        for recipient in shares:
            settings = Object()
            settings._classname = 'ShareConfig'  # pylint: disable=protected-access
            settings.accessMode = recipient.accessMode
            settings.expiration = recipient.expiration
            settings.protectionLevel = recipient.protectionLevel
            settings.invitee = recipient.invitee
            param.shares.append(settings)

    yield param


@contextmanager
def unshare(resource_info, path):
    with share(path, True, True, True) as param:
        logger.info("Unsharing %s: %s", ('folder' if resource_info.isFolder else 'file'), path.reference.as_posix())
        yield param


def add_share_recipient(param, recipient):
    settings = Object()
    settings._classname = 'ShareConfig'  # pylint: disable=protected-access
    settings.accessMode = recipient.access
    if not recipient.access == FileAccessMode.NA:
        settings.expiration = recipient.expiration_date
    invitee = None
    if recipient.type == CollaboratorType.EXT:
        invitee = Object()
        invitee._classname = 'Collaborator'  # pylint: disable=protected-access
        invitee.type = recipient.type
        invitee.email = recipient.account
        if recipient.two_factor:
            settings.protectionLevel = ProtectionLevel.Email
        else:
            settings.protectionLevel = ProtectionLevel.Public
    else:
        invitee = recipient.collaborator
    settings.invitee = invitee
    param.shares.append(settings)


def find_recipients_to_remove(param, path, current_accounts, accounts):
    current_accounts = obtain_current_accounts(param)
    accounts_to_keep, accounts_to_remove = [], []
    if current_accounts:
        for idx, current_account in enumerate(current_accounts):
            if current_account not in accounts:
                accounts_to_keep.append(param.shares[idx])
            else:
                logger.debug('Found recipient to remove: %s', str(current_account))
                accounts_to_remove.append(current_account)
        if not accounts_to_remove:
            logger.info('Share recipients not found: %s', [str(member) for member in accounts])
        else:
            logger.info('Removing share recipients: %s', [str(account) for account in accounts_to_remove])
    else:
        logger.info('Share has no recipients. %s', path.reference.as_posix())
    return accounts_to_keep, accounts_to_remove


def find_recipients_to_add(path, param, current_accounts, valid_recipients):
    current_accounts = obtain_current_accounts(param)
    accounts_to_add = []
    if valid_recipients:
        for recipient in valid_recipients:
            if recipient.account not in current_accounts:
                logger.debug('New share recipient: %s', str(recipient.account))
                accounts_to_add.append(recipient)
            else:
                logger.debug('Share recipient already exists: %s', str(recipient.account))
        if accounts_to_add:
            logger.info('Adding share recipients: %s', [str(recipient.account) for recipient in accounts_to_add])
        else:
            logger.warning('Could not find new share recipients: %s', path.reference.as_posix())
        return accounts_to_add
    logger.warning('Could not find valid share recipients: %s', path.reference.as_posix())
    return accounts_to_add


@contextmanager
def search_collaboration_member(account, cloud_folder_uid):
    param = Object()
    if account.account_type == PortalAccountType.User:
        param.searchType = SearchType.Users
    elif account.account_type == PortalAccountType.Group:
        param.searchType = SearchType.Groups
    else:
        raise ValueError(f'Invalid account type: {account.account_type}')

    param.searchTerm = account.name
    param.resourceUid = cloud_folder_uid
    param.countLimit = 1
    logger.info('Searching for member: %s', str(account))
    yield param


def consume_search_collaboration_response(response, account):
    if not response.objects:
        logger.warning('Could not find results that match your search criteria. %s', {'name': account.name})
    elif len(response.objects) > 1:
        logger.warning('Search returned multiple results. Skipping...')
    else:
        collaborator = response.objects.pop()
        if account.is_local:
            if account.account_type == PortalAccountType.User and collaborator.type == CollaboratorType.LU:
                logger.debug('Found local user account. %s', {'name': account.name})
                return collaborator
            if account.account_type == PortalAccountType.Group and collaborator.type == CollaboratorType.LG:
                logger.debug('Found local group account. %s', {'name': account.name})
                return collaborator
            logger.warning('Expected to find local %s but found domain %s', account.account_type, account.account_type)
        else:
            if account.account_type == PortalAccountType.User and collaborator.type == CollaboratorType.DU:
                logger.debug('Found domain user account. %s', {'name': account.name, 'domain': account.directory})
                return collaborator
            if account.account_type == PortalAccountType.Group and collaborator.type == CollaboratorType.DG:
                logger.debug('Found domain group account. %s', {'name': account.name, 'domain': account.directory})
                return collaborator
            logger.warning('Expected to find domain %s but found local %s', account.account_type, account.account_type)
    return None


def valid_recipient(recipient):
    is_valid = True
    if not recipient.account:
        logger.warning('No account information found. Skipping. %s', {'account': recipient.account})
        is_valid = False
    if not isinstance(recipient.account, (str, UserAccount, GroupAccount)):
        logger.warning('Invalid recipient type. Skipping. %s', {'type': type(recipient.account)})
        is_valid = False
    if recipient.access not in [v for k, v in FileAccessMode.__dict__.items() if not k.startswith('_')]:
        logger.warning('Invalid file access mode. Skipping. %s', {'account': str(recipient.account), 'access': recipient.access})
        is_valid = False
    if recipient.type == CollaboratorType.EXT and not isinstance(recipient.account, str):
        logger.warning('Invalid external recipient type. Skipping. %s', {'expected': 'str', 'actual': type(recipient.account)})
        is_valid = False
    if recipient.type in [CollaboratorType.LU, CollaboratorType.LG] and not recipient.account.is_local:
        logger.warning('Expected local account. Received a domain account. Skipping. %s', {'account': str(recipient.account)})
        is_valid = False
    if recipient.type in [CollaboratorType.DU, CollaboratorType.DG] and recipient.account.is_local:
        logger.warning('Expected domain account. Received a local account. Skipping. %s', {'account': str(recipient.account)})
        is_valid = False
    return is_valid


def obtain_current_accounts(param):
    current_accounts = []
    for collaborator in param.shares:
        if collaborator.invitee.type == CollaboratorType.EXT:
            current_accounts.append(collaborator.invitee.email)
        else:
            current_accounts.append(PortalAccount.from_collaborator(collaborator.invitee))
    return current_accounts


def accept_response(response, reference):
    """
    Check if response contains an error.
    """
    error = {
        "FileWithTheSameNameExist": ResourceExistsError(),
        "DestinationNotExists": PathValidationError(),
        "InvalidName": NameSyntaxError(),
        "ReservedName": ReservedNameError()
    }.get(response, None)
    try:
        if error:
            raise error
    except ResourceExistsError as error:
        logger.info('Resource already exists: a file or folder with this name already exists. %s', {'path': reference})
        raise error
    except PathValidationError as error:
        logger.error('Path validation failed: the specified destination path does not exist. %s', {'path': reference})
        raise error
    except NameSyntaxError as error:
        logger.error('Invalid name: the name contains characters that are not allowed. %s', {'name': reference})
        raise error
    except ReservedNameError as error:
        logger.error('Reserved name error: the name is reserved and cannot be used. %s', {'name': reference})
        raise error



import logging
import asyncio

from .types import Metadata, Block
from .credentials import KeyPair, Bearer
from .crypto import decrypt_key, decrypt_block
from .decompressor import decompress
from ..exceptions.transport import HTTPError
from ..exceptions.direct import (
    UnAuthorized, UnprocessableContent, BlocksNotFoundError, DownloadError, DownloadTimeout, BlockListTimeout,
    DownloadConnectionError, DecryptKeyError, DecryptBlockError, NotFoundError, DecompressBlockError,
    BlockValidationException, BlockListConnectionError, DirectIOError
)


logger = logging.getLogger('cterasdk.direct')


async def retry(coro, retries=3, backoff=1):
    """
    Retry Coroutine.

    :param coroutine coro: Asynchronous Coroutine.
    :param int retries: Retries.
    :param int backoff: Seconds.
    """
    attempts = 0
    while attempts < retries:
        try:
            return await coro()
        except DirectIOError as error:
            attempts = attempts + 1
            if attempts == retries:
                raise error
            wait = backoff * (2 ** (attempts - 1))
            logger.debug('Download of block failed on attempt %s. Retrying in %s seconds...', attempts, wait)
            await asyncio.sleep(wait)


async def get_object(client, file_id, chunk):
    """
    Get Object from a Signed URL.

    :param int file_id: File ID.
    :param cterasdk.direct.types.Chunk chunk: Chunk.
    :returns: Object
    :rtype: bytes
    """
    async def get_object_coro():

        message = (
            f"Downloading block #{chunk.index} "
            f"(offset={chunk.offset}, length={chunk.length})"
        )

        if file_id:
            message += f" for file ID {file_id}"

        error_message, exception = None, None

        logger.debug(message)
        try:
            response = await client.get(chunk.url)
            return await response.read()
        except ConnectionError:
            error_message = 'connection'
            exception = DownloadConnectionError(file_id, chunk)
        except asyncio.TimeoutError:
            error_message = 'timeout'
            exception = DownloadTimeout(file_id, chunk)
        except IOError as error:
            error_message = 'io'
            exception = DownloadError(error, file_id, chunk)
        except HTTPError as error:
            error_message = 'unknown'
            exception = DownloadError(error.error, file_id, chunk)

        error_messages = {
            "connection": "Connection error",
            "timeout": "Timed out",
            "io": "I/O error",
            "unknown": "Unknown error"
        }

        message = (
            f"Failed to download block #{chunk.index} "
            f"(offset={chunk.offset}, length={chunk.length})"
        )
        if file_id:
            message = message + f" for file ID {file_id}"

        message = message + f": {error_messages.get(error_message, 'Unknown error')}."
        logger.error(message)
        raise exception

    return await retry(get_object_coro)


async def decrypt_object(file_id, encrypted_object, encryption_key, chunk):
    """
    Decrypt Encrypted Object.

    :param bytes encrypted_object: Encrypted object.
    :param bytes encryption_key: Encryption key.
    :param cterasdk.direct.types.Chunk chunk: Chunk.
    :returns: Decrypted Object.
    :rtype: bytes
    """
    try:
        return decrypt_block(encrypted_object, encryption_key)
    except DirectIOError:
        logger.error('Failed to decrypt block.')
        raise DecryptBlockError(file_id, chunk)


async def decompress_object(file_id, compressed_object, chunk):
    """
    Decompress Object.

    :param bytes compressed_object: Compressed object.
    :param cterasdk.direct.types.Chunk chunk: Chunk.
    :returns: Decompressed Object.
    :rtype: bytes
    """
    try:
        decompressed_object = decompress(compressed_object)
        if chunk.length != len(decompressed_object):
            logger.error('Expected block length does not match decrypted and decompressed block length.')
            raise BlockValidationException(file_id, chunk)
        return decompressed_object
    except DirectIOError:
        logger.error('Failed to decompress block.')
        raise DecompressBlockError(file_id, chunk)


async def process_chunk(client, file_id, chunk, encryption_key, semaphore):
    """
    Process a Chunk.

    :param cterasdk.clients.clients.AsyncClient client: Asynchronous HTTP Client.
    :param int file_id: File ID.
    :param cterasdk.direct.types.Chunk chunk: Chunk.
    :param str encryption_key: Encryption key.
    :param asyncio.Semaphore semaphore: Semaphore.

    :returns: Block
    :rtype: cterasdk.direct.types.Block
    """
    async def process(client, chunk, encryption_key):
        message = (
            f"Processing block {chunk.index} "
            f"(offset={chunk.offset}, length={chunk.length})"
        )
        if file_id:
            message = message + f" for file ID {file_id}"
        logger.debug(message)
        encrypted_object = await get_object(client, file_id, chunk)
        decrypted_object = await decrypt_object(file_id, encrypted_object, encryption_key, chunk)
        decompressed_object = await decompress_object(file_id, decrypted_object, chunk)
        return Block(file_id, chunk.index, chunk.offset, decompressed_object, chunk.length)

    if semaphore is not None:
        async with semaphore:
            return await process(client, chunk, encryption_key)
    return await process(client, chunk, encryption_key)


async def process_chunks(client, file_id, chunks, encryption_key, semaphore=None):
    """
    Process Chunks Asynchronously.

    :param cterasdk.clients.clients.AsyncClient client: Asynchronous HTTP Client.
    :param int file_id: File ID.
    :param list[cterasdk.direct.types.Chunk] chunks: Chunk.
    :param str encryption_key: Encryption key.
    :param asyncio.Semaphore,optional semaphore: Semaphore.
    :returns: List of futures.
    :rtype: list[asyncio.Task]
    """
    message = [f"Processing {len(chunks)} blocks"]
    if file_id:
        message.append(f"for file ID {file_id}")
    if semaphore:
        message.append(f"using up to {semaphore._value} workers")  # pylint: disable=protected-access
    logger.debug(' '.join(message))
    futures = []
    for chunk in chunks:
        futures.append(asyncio.create_task(process_chunk(client, file_id, chunk, encryption_key, semaphore)))
    return futures


def decrypt_encryption_key(file_id, wrapped_key, secret_access_key):
    """
    Decrypt Encryption Key.

    :param int file_id: File ID.
    :param str wrapped_key: Encryption Key.
    :param str secret_access_key: Secret Access Key.
    :returns: Decrypted Encryption Key.
    :rtype: bytes
    """
    try:
        return decrypt_key(wrapped_key, secret_access_key)
    except DirectIOError:
        logger.error('Failed to decrypt secret key.')
        raise DecryptKeyError(file_id)


def create_authorization_header(credentials):
    """
    Create Authorization Header.

    :param cterasdk.direct.credentials.BaseCredentials credentials: Credentials
    :returns: Authorization header as a dictionary.
    :rtype: dict
    """
    authorization_header = None

    if isinstance(credentials, Bearer):
        logger.debug('Initializing client using bearer token')
        authorization_header = f'Bearer {credentials.bearer}'

    elif isinstance(credentials, KeyPair):
        logger.debug('Initializing client using key pair.')
        authorization_header = f'Bearer {credentials.access_key_id}'

    return {'Authorization': authorization_header}


async def get_chunks(api, credentials, file_id):
    """
    Get Chunks.

    :param cterasdk.clients.clients.AsyncJSON api: Asynchronous JSON Client.
    :param int file_id: File ID.
    :returns: Wrapped key and file chunks.
    :rtype: cterasdk.direct.types.Metadata
    """
    async def get_chunks_coro():
        logger.debug('Listing blocks for file ID: %s', file_id)
        try:
            response = await api.get(f'{file_id}', headers=create_authorization_header(credentials))
            if not response.chunks:
                logger.error('Could not find blocks for file ID: %s.', file_id)
                raise BlocksNotFoundError(file_id)
            return Metadata(file_id, response)
        except HTTPError as error:
            if error.code == 400:
                raise NotFoundError(file_id)
            if error.code == 401:
                raise UnAuthorized(file_id)
            if error.code == 422:
                raise UnprocessableContent(file_id)
            raise error
        except ConnectionError:
            logger.error('Failed to list blocks for file ID: %s due to a connection error.', file_id)
            raise BlockListConnectionError(file_id)
        except asyncio.TimeoutError:
            logger.error('Timed out while listing blocks for file ID: %s.', file_id)
            raise BlockListTimeout(file_id)

    return await retry(get_chunks_coro)



import urllib.parse
from ..common import Object


class URI:
    """A Class Representing a URI"""

    def __init__(self, uri):
        self._uri = urllib.parse.urlparse(uri)

    @property
    def scheme(self):
        return self._uri.scheme

    @property
    def host(self):
        return self._uri.hostname

    @property
    def port(self):
        return self._uri.port or 443

    @property
    def netloc(self):
        return self._uri.netloc

    @property
    def path(self):
        return self._uri.path

    def to_server_object(self):
        param = Object()
        param.scheme = self.scheme
        param.netloc = self.netloc
        param.host = self.host
        param.port = self.port
        param.path = self.path
        param.uri = str(self)
        return param

    @staticmethod
    def instance(scheme, host, port=None, path=None, query=None):
        """
        Create a URI from a list of arguments

        :param str host: Host
        :param int,optional port: Port
        :param str,optional path: Path
        :param dict[str],optional query: Query string parameters
        """
        query = '&'.join([f"{k}={v}" for k, v in query.items()]) if query else ''
        netloc = f"{host}" + (f":{port}" if port else '')
        return URI(urllib.parse.urlunparse(
            (scheme, netloc, path if path else '', '', query, '')
        ))

    def __str__(self):
        return self._uri.geturl()



import logging
from ..exceptions import CTERAException
from ..common.utils import tcp_connect


logger = logging.getLogger('cterasdk.edge')


def test(Edge):
    tcp_connect(Edge.host(), Edge.port())
    logger.debug('Trying to obtain login info.')
    response = Edge.api.get('/nosession/logininfo')
    if response is not None:
        logger.debug('Successfully obtained login info for: %s', response.hostname)
        return response
    raise CTERAException('Could not obtain login info for device (host={host}, port={port}).')



from .base import CTERAException


class RemoteStorageException(CTERAException):
    """
    Base Exception for Remote File Storage

    :ivar str path: Path
    """
    def __init__(self, message, path):
        super().__init__(message)
        self.path = path


class ResourceNotFoundError(RemoteStorageException):

    def __init__(self, path):
        super().__init__('Remote directory not found. Please verify the path and try again.', path)


class NotADirectory(RemoteStorageException):

    def __init__(self, path):
        super().__init__('Target validation error: Resource exists but it is not a directory.', path)


class ResourceExistsError(CTERAException):

    def __init__(self):
        super().__init__('Resource already exists: a file or folder with this name already exists.')


class PathValidationError(CTERAException):

    def __init__(self):
        super().__init__('Path validation failed: the specified destination path does not exist.')


class NameSyntaxError(CTERAException):

    def __init__(self):
        super().__init__('Invalid name: the name contains characters that are not allowed.')


class ReservedNameError(CTERAException):

    def __init__(self):
        super().__init__('Reserved name error: the name is reserved and cannot be used.')


class RestrictedPathError(CTERAException):

    def __init__(self):
        super().__init__('Creating a folder in the specified location is forbidden.')



import logging
import functools

from ..exceptions.session import SessionExpired, NotLoggedIn


logger = logging.getLogger('cterasdk.common')


def authenticated(execute_request):
    @functools.wraps(execute_request)
    def authenticate_then_execute(self, *args, **kwargs):
        if callable(self._authenticator) and self._authenticator(self._builder(args[0])):  # pylint: disable=protected-access
            try:
                return execute_request(self, *args, **kwargs)
            except SessionExpired:
                logger.error('Session expired.')
                self.cookies.clear()
        logger.error('Not logged in.')
        raise NotLoggedIn()
    return authenticate_then_execute



import logging

from ..exceptions import CTERAException
from ..common import Object
from . import enum
from .base_command import BaseCommand


logger = logging.getLogger('cterasdk.edge')


class Syslog(BaseCommand):
    """ Edge Filer Syslog configuration APIs """

    def enable(self, server, port=514, proto=enum.IPProtocol.UDP, min_severity=enum.Severity.INFO):
        """
        Enable Syslog

        :param str server: Server address to send syslog logs
        :param int,optional port: Syslog server communication port, defaults to 514
        :param cterasdk.edge.enum.IPProtocol,optional proto:
         Syslog server communication protocol, defaults to cterasdk.edge.enum.IPProtocol.UDP
        :param cterasdk.edge.enum.Severity,optional min_severity:
         Minimal log severity to fetch, defaults to cterasdk.edge.enum.Severity.INFO
        """
        obj = Object()
        obj.mode = enum.Mode.Enabled
        obj.server = server
        obj.port = port
        obj.proto = proto
        obj.minSeverity = min_severity

        logger.info("Configuring syslog server.")
        self._edge.api.put('/config/logging/syslog', obj)
        logger.info(
            "Syslog server configured. %s",
            {'server': server, 'port': port, 'protocol': proto, 'minSeverity': min_severity}
        )

    def disable(self):
        """ Disable Syslog """
        logger.info("Disabling syslog server.")
        self._edge.api.put('/config/logging/syslog/mode', enum.Mode.Disabled)
        logger.info("Syslog server disabled.")

    def get_configuration(self):
        return self._edge.api.get('/config/logging/syslog')

    def modify(self, server=None, port=None, proto=None, min_severity=None):
        """
        Modify current Syslog configuration. Only configurations that are not None will be changed. Syslog must be enabled

        :param str,optional server: Server address to send syslog logs
        :param int,optional port: Syslog server communication port
        :param cterasdk.edge.enum.IPProtocol,optional proto: Syslog server communication protocol
        :param cterasdk.edge.enum.Severity,optional min_severity: Minimal log severity to fetch
        """
        current_config = self.get_configuration()
        if current_config.mode == enum.Mode.Disabled:
            raise CTERAException("Syslog configuration cannot be modified when disabled")
        if server:
            current_config.server = server
        if port:
            current_config.port = port
        if proto:
            current_config.proto = proto
        if min_severity:
            current_config.minSeverity = min_severity

        logger.info("Updating syslog server configuration.")
        self._edge.api.put('/config/logging/syslog', current_config)
        logger.info(
            "Syslog server configured. %s",
            {
                'server': current_config.server,
                'port': current_config.port,
                'protocol': current_config.proto,
                'minSeverity': current_config.minSeverity
            }
        )



import aiohttp
from ...exceptions.session import SessionExpired


def tracer():

    async def on_request_redirect(session, context, params):
        # pylint: disable=unused-argument
        if params.response.status == 302:
            location = dict(params.response.headers).get('Location', None)
            if location in ['/ServicesPortal/login.html'] or location.startswith(('https://login.microsoftonline.com')):
                raise SessionExpired()

    conf = aiohttp.TraceConfig()
    conf.on_request_redirect.append(on_request_redirect)
    return conf



import copy
from collections.abc import MutableMapping
from aiohttp import TCPConnector, CookieJar, ClientTimeout
import cterasdk.settings
from .tracers import requests, session, postman


class ClientSessionSettings(MutableMapping):

    def __init__(self, *args, **kwargs):
        self._mapping = {
            'connector': {
                '_classname': TCPConnector,
                'ssl': True
            },
            'timeout': {
                '_classname': ClientTimeout,
                'sock_connect': 10,
                'sock_read': 60
            },
            'cookie_jar': {
                '_classname': CookieJar,
                'unsafe': False
            }
        }
        self._mapping.update(dict(*args, **kwargs))

    def update(self, **kwargs):  # pylint: disable=arguments-differ
        for k, v in self._mapping.items():
            attributes = kwargs.get(k, None)
            self._mapping[k] = attributes
            self._mapping[k]['_classname'] = v.get('_classname', None)

    def __getitem__(self, key):
        mapping = copy.deepcopy(self._mapping)
        attributes = mapping.get(key, None)
        new_instance = attributes.pop('_classname', None)
        if new_instance:
            return new_instance(**attributes)
        return attributes

    def __setitem__(self, key, value):
        self._mapping[key] = value

    def __delitem__(self, key):
        del self._mapping[key]

    def __iter__(self):
        return iter(self._mapping)

    def __len__(self):
        return len(self._mapping)

    def __str__(self):
        return str(self._mapping)


class TraceSettings(MutableMapping):

    def __init__(self):
        self._mapping = {
            'trace_configs': [requests.tracer(), session.tracer()]
        }
        if cterasdk.settings.audit.enabled:
            self._mapping['trace_configs'].append(postman.tracer())

    def __getitem__(self, key):
        return self._mapping.get(key, None)

    def __setitem__(self, key, value):  # pylint: disable=useless-parent-delegation
        return super().__setitem__(key, value)

    def __delitem__(self, key):  # pylint: disable=useless-parent-delegation
        return super().__delitem__(key)

    def __len__(self):  # pylint: disable=useless-parent-delegation
        return super().__len__()

    def __iter__(self):
        return iter(self._mapping)



from ...common import Object


class Event(Object):  # pylint: disable=too-many-instance-attributes
    """
    :ivar str type: Object type
    :ivar str guid: Object Unique Identifier (GUID)
    :ivar bool deleted: Deleted
    :ivar str name: Object Name
    :ivar int,optional folder_id: Cloud Drive Folder ID
    :ivar str,optional file_timestamp: File Timestamp
    :ivar str,optional portal_modified_date: Portal Modified Date
    :ivar str,optional virtual_portal_id: Virtual Portal ID
    :ivar int,optional size: Object Size
    :ivar int,optional id: File ID
    :ivar str,optional acl: ACL SDDL
    :ivar str,optional gvsn: GVSN
    :ivar str,optional parent_guid: Parent Object Unique Identifier (GUID)
    """
    def __init__(  # pylint: disable=redefined-builtin, too-many-arguments
            self, type, guid, deleted, name, folder_id=None, modified=None,
            file_timestamp=None, size=None, id=None, acl=None, gvsn=None,
            parent_guid=None, portal_modified_date=None, virtual_portal_id=None):
        super().__init__()
        self.type = type
        self.guid = guid
        self.folder_id = folder_id
        self.deleted = deleted
        self.name = name
        self.file_timestamp = file_timestamp
        self.portal_modified_date = portal_modified_date if portal_modified_date is not None else modified
        self.virtual_portal_id = virtual_portal_id
        self.size = size
        self.id = id
        self.acl = acl
        self.gvsn = gvsn
        self.parent_guid = parent_guid

    @staticmethod
    def from_server_object(server_object):
        return Event(**server_object.__dict__)



from .cmd import Command  # noqa: E402, F401
from .consent import ask  # noqa: E402, F401
from .tempfile import TempfileServices  # noqa: E402, F401
from .version import Version  # noqa: E402, F401
from .iterator import QueryIterator, BaseResponse, FetchResourcesResponse, \
    DefaultResponse, KeyValueQueryIterator, QueryLogsResponse, CursorResponse  # noqa: E402, F401
from .tracker import track, ErrorStatus  # noqa: E402, F401
from .crypto import CryptoServices, X509Certificate, PrivateKey, create_certificate_chain  # noqa: E402, F401



class BaseCredentials:
    """Base Credentials for CTERA Direct IO"""


class KeyPair(BaseCredentials):
    """Access and Secret Key Pair"""

    def __init__(self, access_key_id, secret_access_key):
        self.access_key_id = access_key_id
        self.secret_access_key = secret_access_key


class Bearer(BaseCredentials):
    """Bearer Token"""

    def __init__(self, bearer):
        self.bearer = bearer



import errno


class DirectIOError(IOError):
    """Base Exception for Direct IO Errors"""


class StreamError(DirectIOError):
    """
    Stream Error

    :ivar int offset: Stream offset
    """

    def __init__(self, filename, offset):
        super().__init__(errno.EIO, 'Failed to stream file', filename)
        self.offset = offset


class DirectIOAPIError(DirectIOError):
    """Direct IO API Error"""

    def __init__(self, error, strerror, filename):
        super().__init__(error, strerror, filename)


class NotFoundError(DirectIOAPIError):

    def __init__(self, filename):
        super().__init__(errno.EBADF, 'File not found', filename)


class UnAuthorized(DirectIOAPIError):

    def __init__(self, filename):
        super().__init__(errno.EACCES, 'Unauthorized: You do not have the necessary permissions to access this resource', filename)


class UnprocessableContent(DirectIOAPIError):

    def __init__(self, filename):
        super().__init__(errno.ENOTSUP, 'Not all blocks of the requested file are stored on a storage node set to Direct Mode', filename)


class BlocksNotFoundError(DirectIOAPIError):

    def __init__(self, filename):
        super().__init__(errno.ENODATA, f'Could not find blocks for file ID: {filename}', filename)


class BlockListConnectionError(DirectIOAPIError):

    def __init__(self, filename):
        super().__init__(errno.ENETRESET, f'Failed to list blocks for file ID: {filename} due to a connection error', filename)


class BlockListTimeout(DirectIOAPIError):

    def __init__(self, filename):
        super().__init__(errno.ETIMEDOUT, f'Timed out while listing blocks for file ID: {filename}', filename)


class DecryptKeyError(DirectIOError):

    def __init__(self, filename):
        super().__init__(errno.EIO, 'Failed to decrypt secret key', filename)


class BlockError(DirectIOError):
    """
    Direct IO Block Error

    :ivar cterasdk.direct.exceptions.BlockInfo block: Block info
    """
    def __init__(self, error, strerror, file_id, chunk):
        super().__init__(error, strerror, file_id)
        self.block = BlockInfo(file_id, chunk)


class DownloadError(BlockError):

    def __init__(self, strerror, file_id, chunk):
        super().__init__(errno.EIO, strerror, file_id, chunk)


class DownloadTimeout(BlockError):

    def __init__(self, file_id, chunk):
        super().__init__(errno.ETIMEDOUT, 'Failed to download block. Timed out', file_id, chunk)


class DownloadConnectionError(BlockError):

    def __init__(self, file_id, chunk):
        super().__init__(errno.ENETRESET, 'Failed to download block. Connection error', file_id, chunk)


class DecryptBlockError(BlockError):

    def __init__(self, file_id, chunk):
        super().__init__(errno.EIO, 'Failed to decrypt block', file_id, chunk)


class DecompressBlockError(BlockError):

    def __init__(self, file_id, chunk):
        super().__init__(errno.EIO, 'Failed to decompress block', file_id, chunk)


class BlockValidationException(BlockError):

    def __init__(self, file_id, chunk):
        super().__init__(errno.EIO, 'Expected block length does not match decrypted and decompressed block length', file_id, chunk)


class BlockInfo:
    """
    Block Info

    :ivar int file_id: File ID
    :ivar int number: Block number
    :ivar int offset: Block offset
    :ivar int length: Block length
    """
    def __init__(self, file_id, chunk):
        """
        Initialize an Block Info Object for Direct IO Error Object.

        :param cterasdk.direct.types.Chunk chunk: Chunk.
        """
        self.file_id = file_id
        self.number = chunk.index
        self.offset = chunk.offset
        self.length = chunk.length



import logging
from pathlib import Path


from .base_command import BaseCommand
from ..common import Object
from ..core.enum import Severity, Mode, IPProtocol
from ..exceptions import CTERAException
from ..lib import X509Certificate, PrivateKey
from ..clients.common import MultipartForm


logger = logging.getLogger('cterasdk.core')


class Syslog(BaseCommand):
    """
    Portal Syslog Management APIs
    """

    def import_ca(self, certificate):
        """
        Import the Syslog Server CA certificate

        :param str certificate: Path to the PEM-encoded CA certificate.
        """
        X509Certificate.load_certificate(certificate)
        logging.getLogger('cterasdk.edge').info("Uploading syslog server CA certificate.")
        self._import_secret('/settings/logSettings/syslogConfig/caCertificateUpload', certificate)
        logging.getLogger('cterasdk.edge').info("Uploaded syslog server CA certificate.")

    def import_client_certificate(self, private_key, certificate):
        """
        Import the Syslog Server CA certificate

        :param str private_key: Path to the PEM-encoded private key.
        :param str certificate: Path to the PEM-encoded client certificate.
        """
        PrivateKey.load_private_key(private_key)
        logging.getLogger('cterasdk.edge').info("Uploading syslog server private key.")
        self._import_secret('/settings/logSettings/syslogConfig/clientPrivateKeyUpload', private_key)
        logging.getLogger('cterasdk.edge').info("Uploaded syslog server private key.")

        X509Certificate.load_certificate(certificate)
        logging.getLogger('cterasdk.edge').info("Uploading syslog server client certificate.")
        self._import_secret('/settings/logSettings/syslogConfig/clientCertificateUpload', certificate)
        logging.getLogger('cterasdk.edge').info("Uploaded syslog server client certificate.")

    def _import_secret(self, path, file):
        """
        Import a Syslog Certificate or Private Key

        :param str path: URL Path
        :param str file: File Path
        """
        handle = Path(file)
        with handle.open('rb') as fd:
            form = MultipartForm()
            form.add('name', handle.name)
            form.add('firmware_path', fd, handle.name)
            self._core.api.multipart(path, form)

    def is_enabled(self):
        """
        Check if forwarding log messages over syslog is enabled
        """
        return self._core.api.get('/settings/logsSettings/syslogConfig/mode') == Mode.Enabled

    def get_configuration(self):
        """
        Retrieve the syslog server configuration
        """
        return self._core.api.get('/settings/logsSettings/syslogConfig')

    def enable(self, server, port=514, protocol=IPProtocol.UDP, min_severity=Severity.INFO, ca_cert=None):
        """
        Enable Syslog

        :param str server: Syslog server address
        :param int,optional port: Syslog server port
        :param cterasdk.core.enum.IPProtocol,optional protocol: Syslog server IP protocol
        :param cterasdk.core.enum.Severity,optional min_severity: Minimum Log Severity
        :param str,optional ca_cert: Path to the PEM-encoded CA certificate.
        """
        param = Object()
        param._classname = 'PortalSyslogConfig'  # pylint: disable=protected-access
        param.mode = Mode.Enabled
        param.server = server
        param.minSeverity = min_severity
        param.port = port
        param.protocol = protocol
        param.useClientCertificate = False
        logger.info('Enabling syslog.')
        response = self._core.api.put('/settings/logsSettings/syslogConfig', param)
        if protocol == IPProtocol.TCP and ca_cert is not None:
            self.import_ca(ca_cert)
        logger.info('Syslog enabled.')
        return response

    def modify(self, server=None, port=None, protocol=None, min_severity=None):
        """
        Modify Syslog log forwarding configuration

        :param str server: Syslog server address
        :param int,optional port: Syslog server port
        :param cterasdk.core.enum.IPProtocol,optional protocol: Syslog server IP protocol
        :param cterasdk.core.enum.Severity,optional min_severity: Minimum log severity to forward
        """
        current_config = self.get_configuration()
        if current_config.mode == Mode.Disabled:
            raise CTERAException("Syslog configuration cannot be modified when disabled")
        if server:
            current_config.server = server
        if port:
            current_config.port = port
        if protocol:
            current_config.protocol = protocol
        if min_severity:
            current_config.minSeverity = min_severity

        logger.info("Updating syslog server configuration.")
        self._core.api.put('/settings/logsSettings/syslogConfig', current_config)
        logger.info(
            "Syslog server configured. %s",
            {
                'server': current_config.server,
                'port': current_config.port,
                'protocol': current_config.protocol,
                'minSeverity': current_config.minSeverity
            }
        )

    def disable(self):
        logger.info('Disabling syslog.')
        response = self._core.api.put('/settings/logsSettings/syslogConfig/mode', Mode.Disabled)
        logger.info('Syslog disabled.')
        return response



import logging
from ...cio.common import encode_request_parameter
from ...cio import core as fs
from ...exceptions.io import ResourceNotFoundError, ResourceExistsError, NotADirectory
from ...core import query
from ..enum import CollaboratorType
from ...lib import FetchResourcesResponse


logger = logging.getLogger('cterasdk.core')


def listdir(core, path, depth=None, include_deleted=False, search_criteria=None, limit=None):
    with fs.fetch_resources(path, depth, include_deleted, search_criteria, limit) as param:
        if param.depth > 0:
            return query.iterator(core, '', param, 'fetchResources', callback_response=FetchResourcesResponse)
        return core.api.execute('', 'fetchResources', param)


def exists(core, path):
    present, *_ = metadata(core, path, suppress_error=True)
    return present


def metadata(core, path, suppress_error=False):
    """
    Get item metadata.

    :returns: A tuple indicating if a file exists, and its metadata
    """
    response = listdir(core, path, 0)
    if response.root is None:
        if not suppress_error:
            raise ResourceNotFoundError(path.absolute)
        return False, None
    return True, response.root


def versions(core, path):
    with fs.versions(path):
        return core.api.execute('', 'listSnapshots', path.absolute)


def walk(core, scope, path, include_deleted=False):
    paths = [fs.CorePath.instance(scope, path)]
    while len(paths) > 0:
        path = paths.pop(0)
        entries = listdir(core, path, include_deleted=include_deleted)
        for e in entries:
            if e.isFolder:
                paths.append(fs.CorePath.instance(scope, e))
            yield e


def mkdir(core, path):
    with fs.makedir(path) as param:
        response = core.api.execute('', 'makeCollection', param)
    fs.accept_response(response, path.reference.as_posix())


def makedirs(core, path):
    directories = path.parts
    for i in range(1, len(directories) + 1):
        path = fs.CorePath.instance(path.scope, '/'.join(directories[:i]))
        try:
            mkdir(core, path)
        except ResourceExistsError:
            logger.debug('Resource already exists: %s', path.reference.as_posix())


def rename(core, path, name):
    with fs.rename(path, name) as param:
        return core.api.execute('', 'moveResources', param)


def remove(core, *paths):
    with fs.delete(*paths) as param:
        return core.api.execute('', 'deleteResources', param)


def recover(core, *paths):
    with fs.recover(*paths) as param:
        return core.api.execute('', 'restoreResources', param)


def copy(core, *paths, destination=None):
    with fs.copy(*paths, destination=destination) as param:
        return core.api.execute('', 'copyResources', param)


def move(core, *paths, destination=None):
    with fs.move(*paths, destination=destination) as param:
        return core.api.execute('', 'moveResources', param)


def ensure_directory(core, directory, suppress_error=False):
    present, resource = metadata(core, directory, suppress_error=True)
    if (not present or not resource.isFolder) and not suppress_error:
        raise NotADirectory(directory.absolute)
    return resource.isFolder if present else False, resource


def handle(path):
    """
    Create function to retrieve file handle.

    :param cterasdk.cio.edge.CorePath path: Path to file.
    :returns: Callable function to retrieve file handle.
    :rtype: callable
    """
    def wrapper(core):
        """
        Get file handle.

        :param cterasdk.objects.synchronous.core.Portal core: Portal object.
        """
        with fs.handle(path) as param:
            return core.io.download(param)
    return wrapper


def handle_many(directory, *objects):
    """
    Create function to retrieve zip archive

    :param cterasdk.cio.edge.CorePath directory: Path to directory.
    :param args objects: List of files and folders.
    :returns: Callable function to retrieve file handle.
    :rtype: callable
    """
    def wrapper(core):
        """
        Upload file from metadata and file handle.

        :param cterasdk.objects.synchronous.core.Portal core: Portal object.
        :param str name: File name.
        :param object handle: File handle.
        """
        with fs.handle_many(directory, objects) as param:
            _, resource = ensure_directory(core, directory)
            return core.io.download_zip(str(resource.cloudFolderInfo.uid), encode_request_parameter(param))
    return wrapper


def _validate_destination(core, name, destination):
    is_dir, resource = ensure_directory(core, destination, suppress_error=True)
    if not is_dir:
        is_dir, resource = ensure_directory(core, destination.parent)
        return resource.cloudFolderInfo.uid, destination.name, destination.parent
    return resource.cloudFolderInfo.uid, name, destination


def upload(name, size, destination, fd):
    """
    Create upload function

    :param str name: File name.
    :param cterasdk.cio.core.CorePath destination: Path to directory.
    :param object fd: File handle.
    :returns: Callable function to start the upload.
    :rtype: callable
    """
    def wrapper(core):
        """
        Upload file from metadata and file handle.

        :param cterasdk.objects.synchronous.core.Portal core: Portal object.
        """
        uid, filename, directory = _validate_destination(core, name, destination)
        with fs.upload(core, filename, directory, size, fd) as param:
            return core.io.upload(str(uid), param)
    return wrapper


def _search_collaboration_member(core, account, cloud_folder_uid):
    with fs.search_collaboration_member(account, cloud_folder_uid) as param:
        response = core.api.execute('', 'searchCollaborationMembers', param)
    return fs.consume_search_collaboration_response(response, account)


def remove_share_recipients(core, path, accounts):
    share_info = get_share_info(core, path)
    current_accounts = fs.obtain_current_accounts(share_info)
    accounts_to_keep, accounts_to_remove = fs.find_recipients_to_remove(share_info, path, current_accounts, accounts)
    if accounts_to_remove:
        with fs.share(path, share_info.teamProject, share_info.allowReshare, share_info.shouldSync, accounts_to_keep) as param:
            core.api.execute('', 'shareResource', param)
    return accounts_to_remove


def get_share_info(core, path):
    with fs.share_info(path) as param:
        return core.api.execute('', 'listShares', param)


def share(core, path, recipients, as_project, allow_reshare, allow_sync):
    valid_recipients = _obtain_valid_recipients(core, path, recipients)
    if valid_recipients:
        with fs.share(path, as_project, allow_reshare, allow_sync) as param:
            for recipient in valid_recipients:
                fs.add_share_recipient(param, recipient)
            logger.info('Sharing: %s with: %s', path.reference.as_posix(), [str(recipient) for recipient in valid_recipients])
            core.api.execute('', 'shareResource', param)
            return valid_recipients
    logger.warning('Resource not shared. Could not find valid recipients: %s', path.reference.as_posix())
    return valid_recipients


def add_share_recipients(core, path, recipients):
    share_info = get_share_info(core, path)
    current_accounts = fs.obtain_current_accounts(share_info)
    valid_recipients = _obtain_valid_recipients(core, path, recipients)
    with fs.share(path, share_info.teamProject, share_info.allowReshare, share_info.shouldSync, share_info.shares) as param:
        accounts_to_add = fs.find_recipients_to_add(path, share_info, current_accounts, valid_recipients)
        if accounts_to_add:
            for recipient in accounts_to_add:
                fs.add_share_recipient(param, recipient)
            core.api.execute('', 'shareResource', param)
    return valid_recipients


def _obtain_valid_recipients(core, path, recipients):
    _, resource_info = metadata(core, path)
    valid_recipients = []
    for recipient in filter(fs.valid_recipient, recipients):
        if not recipient.type == CollaboratorType.EXT:
            collaborator = _search_collaboration_member(core, recipient.account, resource_info.cloudFolderInfo.uid)
            if collaborator:
                recipient.collaborator = collaborator
                valid_recipients.append(recipient)
        else:
            valid_recipients.append(recipient)
    return valid_recipients


def unshare(core, path):
    _, resource_info = metadata(core, path)
    with fs.unshare(resource_info, path) as param:
        return core.api.execute('', 'shareResource', param)


def public_link(core, path, access, expire_in):
    with fs.public_link(path, access, expire_in) as param:
        response = core.api.execute('', 'createShare', param)
    return response.publicLink



import re
import os
import logging
import functools

from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization import Encoding, PrivateFormat, PublicFormat, NoEncryption, load_pem_private_key

from ..exceptions import CTERAException
from .storage import synfs, commonfs


logger = logging.getLogger('cterasdk.crypto')


class RSAKeyPair:

    def __init__(self, private_key, public_key):
        self._private_key = private_key
        self._public_key = public_key

    @property
    def public_key(self):
        return self._public_key.public_bytes(Encoding.OpenSSH, PublicFormat.OpenSSH)

    @property
    def private_key(self):
        return self._private_key.private_bytes(Encoding.PEM, PrivateFormat.OpenSSH, NoEncryption())

    def save(self, dirpath, key_filename):
        logger.info('Saving private key.')
        path = synfs.write(dirpath, f'{key_filename}.pem', self.private_key)
        os.chmod(path, 0o600)
        logger.info('Saved private key. %s', {'filepath': path, 'format': 'PEM'})

        logger.info('Saving public key.')
        path = synfs.write(dirpath, f'{key_filename}.pub', self.public_key)
        logger.info('Saved public key. %s', {'filepath': path, 'format': 'OpenSSH'})


class CryptoServices:

    @staticmethod
    def generate_rsa_key_pair(exponent=65537, key_size=2048):
        private_key = rsa.generate_private_key(public_exponent=exponent, key_size=key_size)
        public_key = private_key.public_key()
        return RSAKeyPair(private_key, public_key)

    @staticmethod
    def generate_and_save_key_pair(key_filename, exponent=65537, key_size=2048, dirpath=None):
        key_pair = CryptoServices.generate_rsa_key_pair(exponent, key_size)
        if not dirpath:
            dirpath = commonfs.downloads()
        key_pair.save(dirpath, key_filename)
        return key_pair.public_key.decode('utf-8')


def create_certificate_chain(*certificates):
    return sorted(certificates, key=functools.cmp_to_key(compare_certificates))


def compare_certificates(a, b):
    if a.subject == b.issuer:
        return 1
    if b.subject == a.issuer:
        return -1
    return 0


class PrivateKey:

    def __init__(self, private_key):
        self.private_key = private_key

    @property
    def pem_data(self):
        return self.private_key.private_bytes(Encoding.PEM, PrivateFormat.TraditionalOpenSSL, NoEncryption())

    @staticmethod
    def from_file(path, password=None):
        with open(path, 'r', encoding='utf-8') as f:
            data = f.read()
        return PrivateKey.from_string(data, password)

    @staticmethod
    def from_string(data, password=None):
        return PrivateKey.from_bytes(data.encode('utf-8'), password)

    @staticmethod
    def from_bytes(data, password=None):
        return PrivateKey(load_pem_private_key(data, password))

    @staticmethod
    def load_private_key(key, password=None):
        try:
            if isinstance(key, bytes):
                return PrivateKey.from_bytes(key, password)

            if commonfs.exists(key):
                return PrivateKey.from_file(key, password)
            return PrivateKey.from_string(key, password)
        except ValueError as error:
            logger.error('Failed loading private key.')
            raise CTERAException('Failed loading private key.') from error


class X509Certificate:

    def __init__(self, certificate):
        self.certificate = certificate

    @property
    def sha1_fingerprint(self):
        hexstr = self.certificate.fingerprint(hashes.SHA1()).hex()
        return ':'.join([a + b for a, b in zip(hexstr[::2], hexstr[1::2])])

    @property
    def issuer(self):
        return X509Certificate._parse_common_name(self.certificate.issuer.rdns[-1].rfc4514_string())

    @property
    def subject(self):
        return X509Certificate._parse_common_name(self.certificate.subject.rdns[-1].rfc4514_string())

    @property
    def is_root(self):
        return self.issuer == self.subject

    @property
    def pem_data(self):
        return self.certificate.public_bytes(Encoding.PEM)

    @staticmethod
    def _parse_common_name(common_name):
        match = re.search(r'(?<=CN=).+$', common_name)
        return match.group(0) if match else None

    @staticmethod
    def from_file(path):
        with open(path, 'r', encoding='utf-8') as f:
            data = f.read()
        return X509Certificate.from_string(data)

    @staticmethod
    def from_string(data):
        return X509Certificate.from_bytes(data.encode('utf-8'))

    @staticmethod
    def from_bytes(data):
        return X509Certificate(x509.load_pem_x509_certificate(data))

    @staticmethod
    def load_certificate(cert):
        try:
            if isinstance(cert, bytes):
                return X509Certificate.from_bytes(cert)

            if commonfs.exists(cert):
                return X509Certificate.from_file(cert)
            return X509Certificate.from_string(cert)
        except ValueError as error:
            logger.error('Failed loading certificate.')
            raise CTERAException('Failed loading certificate.') from error

    def __str__(self):
        return str(
            dict(
                issued_to=self.subject,
                issued_by=self.issuer
            )
        )



import logging
from .base import BaseSession, BaseUser
from .types import Product

from ...common import Object


class Connection(Object):
    """Connection"""

    def __init__(self, remote, source=None):
        super().__init__()
        self.remote = remote
        if source:
            self.source = source


class LocalUser(BaseUser):
    """Local User"""


class RemoteUser(BaseUser):
    """Remote User"""

    def __init__(self, name, domain, tenant):
        super().__init__(name, domain)
        self.tenant = tenant


class Session(BaseSession):

    def __init__(self, address):
        super().__init__(address, Product.Edge)
        self.connection = None

    def _start_session(self, session):
        logging.getLogger('cterasdk.edge').debug('Starting Session.')
        username = session.api.get('/currentuser').username
        software_version = session.api.get('/status/device/runningFirmware')
        self._update_session(username, software_version)

    async def _async_start_session(self, session):
        logging.getLogger('cterasdk.edge').debug('Starting Session.')
        user = session.api.get('/currentuser')
        software = session.api.get('/status/device/runningFirmware')
        self._update_session((await user).username, await software)

    def _update_session(self, username, software_version):
        self._update_account(LocalUser(username))
        self._update_software_version(software_version)
        self.connection = Connection(False)

    def start_remote_session(self, session):
        logging.getLogger('cterasdk.edge').debug('Starting Remote Session.')
        self._update_account(RemoteUser(session.account.name, session.account.domain, session.account.tenant))
        self.connection = Connection(True, session.address)

    def _stop_session(self):
        logging.getLogger('cterasdk.edge').debug('Stopping Session.')
        self.connection = None



import logging
from ..common.utils import tcp_connect


logger = logging.getLogger('cterasdk.core')


def test(Portal):
    tcp_connect(Portal.host(), Portal.port())
    logger.debug('Trying to obtain Portal public info.')
    response = Portal.public_info()
    logger.debug('Successfully obtained Portal public info. %s', {'version': response.version, 'name': response.name})
    return response



import asyncio
import logging


from .types import Event
from .base_command import BaseCommand
from ...common import Object
from ...lib import CursorResponse
from ...exceptions.transport import HTTPError
from ...exceptions.notifications import NotificationsError


logger = logging.getLogger('cterasdk.notifications')


class Notifications(BaseCommand):
    """CTERA Portal Notification Service APIs"""

    def __init__(self, core):
        super().__init__(core)
        self.service = Service(core)

    async def get(self, cloudfolders=None, cursor=None, max_results=None):
        """
        List Changes.

        :param list[CloudFSFolderFindingHelper],optional cloudfolders: List of Cloud Drive folders, defaults to all cloud drive folders.
        :param str,optional cursor: Cursor
        :param int,optional max_results: Limit max results, defaults to 2000.

        :returns: An asynchronous iterator
        :rtype: cterasdk.asynchronous.core.iterator.CursorAsyncIterator
        :raises: cterasdk.exceptions.NotificationsError
        """
        param = await self._create_parameter(cloudfolders, cursor)
        param.max_results = max_results if max_results is not None else 2000
        logger.debug('Listing updates.')
        response = await self._core.v2.api.post('/metadata/list', param)
        if response is not None:
            return CursorResponse(response)
        logger.error('An error occurred while trying to retrieve notifications.')
        raise NotificationsError(cloudfolders, cursor)

    async def _create_parameter(self, cloudfolders, cursor):
        param = Object()
        param.folder_ids = []
        param.cursor = cursor
        if cloudfolders:
            if all(isinstance(cloudfolder, int) for cloudfolder in cloudfolders):
                param.folder_ids = cloudfolders
            else:
                for cloudfolder in cloudfolders:
                    async for cloudfolder in await self._core.cloudfs.drives.find(cloudfolder.name, cloudfolder.owner, include=['uid']):
                        param.folder_ids.append(cloudfolder.uid)
        return param

    async def changes(self, cursor, cloudfolders=None, timeout=None):
        """
        Check for Changes.

        :param str cursor: Cursor
        :param list[CloudFSFolderFindingHelper],optional cloudfolders: List of Cloud Drive folders, defaults to all cloud drive folders.
        :param int,optional timeout: Timeout

        :returns: ``True`` if changes are available for this ``cursor``, ``False`` otherwise
        :rtype: bool
        """
        param = Object()
        param = await self._create_parameter(cloudfolders, cursor)
        param.timeout = timeout if timeout else 10000
        logger.debug('Checking for updates. %s', {'timeout': param.timeout})
        return (await self._core.v2.api.post('/metadata/longpoll', param)).changes

    async def ancestors(self, descendant):
        """
        Get Ancestors.

        :param cterasdk.asynchronous.core.types.Event descendant: Event
        :returns: Sorted List of Ancestors
        :rtype: list[cterasdk.common.object.Object]
        """
        param = Object()
        param.folder_id = descendant.folder_id
        param.guid = descendant.guid
        logger.debug('Getting ancestors. %s', {'guid': param.guid, 'folder_id': param.folder_id})
        try:
            return await self._core.v2.api.post('/metadata/ancestors', param)
        except HTTPError:
            logger.error('Could not retrieve ancestors. %s', {'folder_id': param.folder_id, 'guid': param.guid})
            raise


class Service(BaseCommand):
    """Change Notification Service"""

    def __init__(self, core):
        super().__init__(core)
        self._promises = []

    def run(self, client_queue, save_cursor, *, cloudfolders=None, cursor=None):
        """
        Start Service.

        :param asyncio.Queue client_queue: Queue.
        :param callback save_cursor: Asynchronous callback function to persist the cursor.
        :param list[CloudFSFolderFindingHelper] cloudfolders: List of Cloud Drive folders.
        :param str,optional cursor: Cursor.
        """
        server_queue = asyncio.Queue(maxsize=3)
        retrieve = asyncio.create_task(retrieve_events(server_queue, self._core, cloudfolders, cursor))
        forward = asyncio.create_task(forward_events(server_queue, client_queue, save_cursor))
        self._promises = [retrieve, forward]

    async def stop(self):
        """Stop Service"""
        for promise in self._promises:
            promise.cancel()
            try:
                await promise
            except asyncio.CancelledError:
                pass
        self._promises.clear()


async def retrieve_events(server_queue, core, cloudfolders, cursor):
    """
    Retrieval Service.

    :param asyncio.Queue server_queue: Queue.
    :param cterasdk.objects.data.DataServices core: Data Services object.
    :param list[CloudFSFolderFindingHelper] cloudfolders: List of Cloud Drive folders.
    :param str cursor: Cursor
    """
    logger.debug('Event Retrieval Service.')
    last_response = LastResponse(cursor)
    try:
        while True:
            try:
                if last_response.cursor is None or last_response.more or \
                        await core.notifications.changes(last_response.cursor, cloudfolders):
                    response = await core.notifications.get(cloudfolders, last_response.cursor)
                    if response.objects:
                        await server_queue.put(response)
                    last_response = response
            except ConnectionError as error:
                await on_connection_error(error)
            except TimeoutError:
                logger.debug('Request timed out. Retrying.')
    except asyncio.CancelledError:
        logger.debug('Cancelling Event Retrieval.')


async def forward_events(server_queue, client_queue, save_cursor):
    """
    Change Notification Service.

    :param asyncio.Queue server_queue: Server queue.
    :param asyncio.Queue client_queue: Client queue.
    :param callback save_cursor: Callback function to persist the cursor.
    """
    logger.debug('Event Forwarder Service.')
    try:
        while True:
            batch = await server_queue.get()
            await enqueue_events(batch.objects, client_queue)
            await process_events(client_queue)
            await persist_cursor(save_cursor, batch.cursor)
    except asyncio.CancelledError:
        logger.debug('Cancelling Event Forwarding.')


async def enqueue_events(events, queue):
    """
    Enqueue Events.

    :param asyncio.Queue queue: Event Queue.
    :param cterasdk.asynchronous.core.iterator.CursorAsyncIterator events: Event Iterator.
    """
    for event in events:
        logger.debug('Enqueuing Event.')
        await queue.put(Event.from_server_object(event))
        logger.debug('Enqueued Event.')


async def process_events(queue):
    """
    Process Events.

    :param asyncio.Queue queue: Queue.
    """
    logger.debug('Joining Queue.')
    await queue.join()
    logger.debug('Completed Processing.')


async def persist_cursor(save_cursor, cursor):
    """
    Persist Cursor.

    :param callback save_cursor: Asynchronous callback function to persist the cursor.
    :param str cursor: Cursor
    """
    logger.debug("Persisting Cursor. Calling function: '%s'", save_cursor)
    try:
        await save_cursor(cursor)
        logger.debug("Called Persist Cursor Function.")
    except Exception:  # pylint: disable=broad-exception-caught
        logger.error("An error occurred while trying to persist cursor. Function: '%s'", save_cursor)


async def on_connection_error(error):
    seconds = 5
    logger.error('Connection error. Reason: %s.', str(error))
    logger.debug("Retrying in %s seconds.", seconds)
    await asyncio.sleep(seconds)


class LastResponse:

    def __init__(self, cursor):
        self._cursor = cursor

    @property
    def more(self):
        return False

    @property
    def objects(self):
        return []

    @property
    def cursor(self):
        return self._cursor



import logging

from abc import abstractmethod


logger = logging.getLogger('cterasdk.common')


class BaseAsyncIterator:
    """Abstract Asynchronous Iterator"""

    def __init__(self, callback, parameter):
        self._callback = callback
        self._parameter = parameter
        self._more = True
        self._objects = []

    def __aiter__(self):
        return self

    async def __anext__(self):
        if self._objects:
            return self.object
        if self._more:
            self._more, page = await self.page()
            self._objects.extend(page)
            if self._objects:
                return self.object
        logger.debug('Stopping iteration.')
        raise StopAsyncIteration

    @property
    def object(self):
        return self._objects.pop(0)

    @abstractmethod
    async def page(self):
        raise NotImplementedError("Subclass must implemenet the 'page' function")


class QueryAsyncIterator(BaseAsyncIterator):
    """ Asynchronous Objects Iterator """

    async def page(self):
        response = await self._callback(self._parameter)
        self._parameter.increment()
        return response.more, response.objects


class CursorAsyncIterator(BaseAsyncIterator):

    def __init__(self, callback, parameter):
        super().__init__(callback, parameter)
        self._cursor = None

    @property
    def cursor(self):
        return self._parameter.cursor

    async def page(self):
        response = await self._callback(self._parameter)
        self._parameter.cursor = response.cursor
        return response.more, response.objects



import cterasdk.settings
from ..services import AsyncManagement
from ..endpoints import EndpointBuilder
from ...clients import clients
from .. import authenticators
from ...lib.session.edge import Session
from ...asynchronous.edge import login, files


class Clients:

    def __init__(self, edge):
        self.api = edge.default.clone(clients.AsyncAPI, EndpointBuilder.new(edge.base, '/admingui/api'))
        self.io = IO(edge)


class IO:

    def __init__(self, edge):
        self._edge = edge
        self._webdav = edge.default.clone(clients.AsyncWebDAV, EndpointBuilder.new(edge.base, '/localFiles'))

    @property
    def download(self):
        return self._webdav.download

    @property
    def download_zip(self):
        return self._edge.default.form_data  # pylint: disable=protected-access

    @property
    def upload(self):
        return self._edge.default.form_data  # pylint: disable=protected-access

    @property
    def propfind(self):
        return self._webdav.propfind

    @property
    def mkdir(self):
        return self._webdav.mkcol

    @property
    def copy(self):
        return self._webdav.copy

    @property
    def move(self):
        return self._webdav.move

    @property
    def delete(self):
        return self._webdav.delete


class AsyncEdge(AsyncManagement):

    def __init__(self, host=None, port=None, https=True, *, base=None):
        super().__init__(host, port, https, base, cterasdk.settings.edge.asyn.settings)
        self._ctera_session = Session(self.host())
        self._ctera_clients = Clients(self)
        self.files = files.FileBrowser(self)

    @property
    def v1(self):
        return self.clients.v1

    @property
    def api(self):
        return self.clients.api

    @property
    def io(self):
        return self.clients.io

    @property
    def _login_object(self):
        return login.Login(self)

    def _authenticator(self, url):
        return authenticators.edge(self.session(), url)



import logging

from .base_command import BaseCommand
from ...exceptions import CTERAException


logger = logging.getLogger('cterasdk.edge')


class Login(BaseCommand):
    """
    CTERA Edge Filer Login APIs
    """

    async def login(self, username, password):
        host = self._edge.host()
        try:
            await self._edge.api.form_data('/login', {'username': username, 'password': password})
            logger.info("User logged in. %s", {'host': host, 'user': username})
        except CTERAException:
            logger.error("Login failed. %s", {'host': host, 'user': username})
            raise

    async def logout(self):
        """
        Log out of the portal
        """
        host = self._edge.host()
        user = self._edge.session().account.name
        try:
            await self._edge.api.form_data('/logout', {'foo': 'bar'})
            logger.info("User logged out. %s", {'host': host, 'user': user})
        except CTERAException:
            logger.error("Logout failed. %s", {'host': host, 'user': user})
            raise



from .base_command import BaseCommand
from ...common import union, Object
from ...exceptions import ObjectNotFoundException
from ...exceptions.session import ContextError


class Users(BaseCommand):
    """
    CTERA Portal User Management APIs
    """

    default = ['name', 'uid']

    async def get(self, user_account, include=None):
        """
        Get a user account

        :param cterasdk.core.types.UserAccount user_account: User account, including the user directory and user name
        :param list[str] include: List of fields to retrieve, defaults to ['name']

        :return: The user account, including the requested fields
        """
        baseurl = f'/users/{user_account.name}' if user_account.is_local \
            else f'/domains/{user_account.directory}/adUsers/{user_account.name}'
        include = union(include or [], Users.default)
        include = ['/' + attr for attr in include]
        user_object = await self._core.v1.api.get_multi(baseurl, include)
        if user_object.name is None:
            raise ObjectNotFoundException(baseurl)
        return user_object

    async def generate_ticket(self, username, tenant):
        """
        Create an SSO ticket for a user.

        :param str username: Username
        :param str portal: Tenant
        """
        if self.session().in_tenant_context():
            raise ContextError('Browse the Global Administration to invoke this API')
        param = Object()
        param.username = username
        param.portal = tenant
        return await self._core.v1.api.execute('', 'getSessionToken', param)
